#ifndef MSP_CORE_VARIANT_H_
#define MSP_CORE_VARIANT_H_

#include <stdexcept>
#include <typeinfo>
#include "meta.h"

namespace Msp {

class type_mismatch: public std::runtime_error
{
public:
	type_mismatch(const std::type_info &, const std::type_info &);
	~type_mismatch() throw() { }
};


class Variant
{
private:
	struct StoreBase
	{
		virtual ~StoreBase() { }

		virtual const std::type_info &type_id() const = 0;
		virtual StoreBase *clone() const = 0;
	};

	template<typename T>
	struct Store: public StoreBase
	{
		T data;

		Store(T d): data(d) { }

		virtual const std::type_info &type_id() const { return typeid(T); }
		virtual StoreBase *clone() const { return new Store<T>(data); }
	};

	StoreBase *store;

public:
	Variant(): store(0) { }
	template<typename T>
	Variant(const T &v): store(new Store<typename RemoveConst<T>::Type>(v)) { }
	Variant(const Variant &v): store(v.store ? v.store->clone() : 0) { }
	~Variant() { delete store; }

	template<typename T>
	Variant &operator=(const T &v)
	{
		delete store;
		store = new Store<typename RemoveConst<T>::Type>(v);
		return *this;
	}

	Variant &operator=(const Variant &v)
	{
		delete store;
		store = (v.store ? v.store->clone() : 0);
		return *this;
	}

private:
	template<typename T>
	Store<typename RemoveConst<T>::Type> *get_typed_store() const
	{
		typedef typename RemoveConst<T>::Type NCT;
		Store<NCT> *s = dynamic_cast<Store<NCT> *>(store);
		if(!s)
			throw type_mismatch(typeid(T), (store ? store->type_id() : typeid(void)));
		return s;
	}

public:
	template<typename T>
	T &value()
	{
		return get_typed_store<T>()->data;
	}

	template<typename T>
	const T &value() const
	{
		return get_typed_store<T>()->data;
	}

	template<typename T>
	bool check_type() const
	{
		return dynamic_cast<Store<typename RemoveConst<T>::Type> *>(store)!=0;
	}

	template<typename T>
	operator T() const
	{ return value<T>(); }
};

} // namespace Msp

#endif
